#include <catch2/catch_test_macros.hpp>
#include <catch2/matchers/catch_matchers_all.hpp>

#include <MeshDataBaseForTests.hpp>
#include <mesh/CartesianMeshBuilder.hpp>
#include <mesh/Connectivity.hpp>
#include <mesh/ConnectivityUtils.hpp>
#include <mesh/ItemValue.hpp>
#include <mesh/ItemValueUtils.hpp>
#include <mesh/Mesh.hpp>
#include <mesh/MeshFaceBoundary.hpp>
#include <mesh/MeshVariant.hpp>
#include <mesh/NamedBoundaryDescriptor.hpp>
#include <mesh/StencilManager.hpp>
#include <utils/Messenger.hpp>

#include <NbGhostLayersTester.hpp>

// clazy:excludeall=non-pod-global-static

TEST_CASE("StencilBuilder node2cell", "[mesh]")
{
  auto is_valid = []<ItemType source_item_type, ItemType connecting_item_type>(const auto& connectivity,
                                                                               const auto& stencil_array,
                                                                               const size_t number_of_layers) {
    constexpr size_t Dimension = std::decay_t<decltype(connectivity)>::Dimension;

    auto cell_to_connecting_item_matrix =
      connectivity.template getItemToItemMatrix<ItemType::cell, connecting_item_type>();

    auto source_item_to_connecting_item_matrix =
      [](const auto& given_connectivity) -> ItemToItemMatrix<source_item_type, connecting_item_type> {
      if constexpr (ItemTypeId<Dimension>::itemTId(source_item_type) ==
                    ItemTypeId<Dimension>::itemTId(connecting_item_type)) {
        return ConnectivityMatrix{};
      } else {
        return given_connectivity.template getItemToItemMatrix<source_item_type, connecting_item_type>();
      }
    }(connectivity);

    auto connecting_to_cell_matrix = connectivity.template getItemToItemMatrix<connecting_item_type, ItemType::cell>();
    auto cell_is_owned             = connectivity.cellIsOwned();
    auto cell_number               = connectivity.cellNumber();
    auto source_item_is_owned      = connectivity.template isOwned<source_item_type>();

    using SourceItemId     = ItemIdT<source_item_type>;
    using ConnectingItemId = ItemIdT<connecting_item_type>;

    for (SourceItemId source_item_id = 0; source_item_id < connectivity.template numberOf<source_item_type>();
         ++source_item_id) {
      if (source_item_is_owned[source_item_id]) {
        std::vector<CellId> expected_stencil;

        std::set<CellId> marked_cell_set;
        if constexpr (source_item_type == ItemType::cell) {
          marked_cell_set.insert(source_item_id);
        }

        std::set<ConnectingItemId> marked_connecting_item_set;
        std::set<ConnectingItemId> layer_connecting_item_set;
        {
          if constexpr (ItemTypeId<Dimension>::itemTId(source_item_type) ==
                        ItemTypeId<Dimension>::itemTId(connecting_item_type)) {
            ConnectingItemId connecting_id =
              ConnectingItemId{static_cast<typename SourceItemId::base_type>(source_item_id)};
            layer_connecting_item_set.insert(connecting_id);
            marked_connecting_item_set.insert(connecting_id);
          } else {
            auto source_item_to_connecting_item_list = source_item_to_connecting_item_matrix[source_item_id];
            for (size_t i_connecting_item = 0; i_connecting_item < source_item_to_connecting_item_list.size();
                 ++i_connecting_item) {
              const ConnectingItemId connecting_item_id = source_item_to_connecting_item_list[i_connecting_item];
              layer_connecting_item_set.insert(connecting_item_id);
              marked_connecting_item_set.insert(connecting_item_id);
            }
          }
        }

        for (size_t i_layer = 0; i_layer < number_of_layers; ++i_layer) {
          std::set<CellId, std::function<bool(CellId, CellId)>> cell_set(
            [=](CellId cell_0, CellId cell_1) { return cell_number[cell_0] < cell_number[cell_1]; });

          for (auto&& connecting_item_id : layer_connecting_item_set) {
            auto connecting_item_to_cell_list = connecting_to_cell_matrix[connecting_item_id];
            for (size_t i_connecting_item_of_cell = 0; i_connecting_item_of_cell < connecting_item_to_cell_list.size();
                 ++i_connecting_item_of_cell) {
              const CellId connecting_item_id_of_cell = connecting_item_to_cell_list[i_connecting_item_of_cell];
              if (not marked_cell_set.contains(connecting_item_id_of_cell)) {
                cell_set.insert(connecting_item_id_of_cell);
                marked_cell_set.insert(connecting_item_id_of_cell);
              }
            }
          }

          layer_connecting_item_set.clear();
          for (auto layer_cell_id : cell_set) {
            auto cell_to_connecting_item_list = cell_to_connecting_item_matrix[layer_cell_id];
            for (size_t i_connecting_item = 0; i_connecting_item < cell_to_connecting_item_list.size();
                 ++i_connecting_item) {
              const ConnectingItemId connecting_item_id = cell_to_connecting_item_list[i_connecting_item];
              if (not marked_connecting_item_set.contains(connecting_item_id)) {
                layer_connecting_item_set.insert(connecting_item_id);
                marked_connecting_item_set.insert(connecting_item_id);
              }
            }
          }

          for (auto&& set_cell_id : cell_set) {
            expected_stencil.push_back(set_cell_id);
          }
        }

        auto cell_stencil = stencil_array[source_item_id];

        auto i_set_cell = expected_stencil.begin();
        if (cell_stencil.size() != expected_stencil.size()) {
          return false;
        }

        for (size_t index = 0; index < cell_stencil.size(); ++index, ++i_set_cell) {
          if (*i_set_cell != cell_stencil[index]) {
            return false;
          }
        }
      }
    }
    return true;
  };

  SECTION("inner stencil")
  {
    SECTION("1 layer")
    {
      SECTION("1D")
      {
        SECTION("cartesian")
        {
          const auto& mesh = *MeshDataBaseForTests::get().unordered1DMesh()->get<Mesh<1>>();

          const Connectivity<1>& connectivity = mesh.connectivity();

          REQUIRE(
            is_valid.template
            operator()<ItemType::node,
                       ItemType::node>(connectivity,
                                       StencilManager::instance()
                                         .getNodeToCellStencilArray(connectivity,
                                                                    StencilDescriptor{1, StencilDescriptor::
                                                                                           ConnectionType::by_nodes}),
                                       1));

          REQUIRE(
            is_valid.template
            operator()<ItemType::node,
                       ItemType::edge>(connectivity,
                                       StencilManager::instance()
                                         .getNodeToCellStencilArray(connectivity,
                                                                    StencilDescriptor{1, StencilDescriptor::
                                                                                           ConnectionType::by_edges}),
                                       1));

          REQUIRE(
            is_valid.template
            operator()<ItemType::node,
                       ItemType::face>(connectivity,
                                       StencilManager::instance()
                                         .getNodeToCellStencilArray(connectivity,
                                                                    StencilDescriptor{1, StencilDescriptor::
                                                                                           ConnectionType::by_faces}),
                                       1));
        }

        SECTION("unordered")
        {
          const auto& mesh = *MeshDataBaseForTests::get().unordered1DMesh()->get<Mesh<1>>();

          const Connectivity<1>& connectivity = mesh.connectivity();

          REQUIRE(
            is_valid.template
            operator()<ItemType::node,
                       ItemType::node>(connectivity,
                                       StencilManager::instance()
                                         .getNodeToCellStencilArray(connectivity,
                                                                    StencilDescriptor{1, StencilDescriptor::
                                                                                           ConnectionType::by_nodes}),
                                       1));

          REQUIRE(
            is_valid.template
            operator()<ItemType::node,
                       ItemType::edge>(connectivity,
                                       StencilManager::instance()
                                         .getNodeToCellStencilArray(connectivity,
                                                                    StencilDescriptor{1, StencilDescriptor::
                                                                                           ConnectionType::by_edges}),
                                       1));

          REQUIRE(
            is_valid.template
            operator()<ItemType::node,
                       ItemType::face>(connectivity,
                                       StencilManager::instance()
                                         .getNodeToCellStencilArray(connectivity,
                                                                    StencilDescriptor{1, StencilDescriptor::
                                                                                           ConnectionType::by_faces}),
                                       1));
        }
      }

      SECTION("2D")
      {
        SECTION("cartesian")
        {
          const auto& mesh = *MeshDataBaseForTests::get().cartesian2DMesh()->get<Mesh<2>>();

          const Connectivity<2>& connectivity = mesh.connectivity();
          REQUIRE(
            is_valid.template
            operator()<ItemType::node,
                       ItemType::node>(connectivity,
                                       StencilManager::instance()
                                         .getNodeToCellStencilArray(connectivity,
                                                                    StencilDescriptor{1, StencilDescriptor::
                                                                                           ConnectionType::by_nodes}),
                                       1));

          REQUIRE(
            is_valid.template
            operator()<ItemType::node,
                       ItemType::edge>(connectivity,
                                       StencilManager::instance()
                                         .getNodeToCellStencilArray(connectivity,
                                                                    StencilDescriptor{1, StencilDescriptor::
                                                                                           ConnectionType::by_edges}),
                                       1));

          REQUIRE(
            is_valid.template
            operator()<ItemType::node,
                       ItemType::face>(connectivity,
                                       StencilManager::instance()
                                         .getNodeToCellStencilArray(connectivity,
                                                                    StencilDescriptor{1, StencilDescriptor::
                                                                                           ConnectionType::by_faces}),
                                       1));
        }

        SECTION("hybrid")
        {
          const auto& mesh = *MeshDataBaseForTests::get().hybrid2DMesh()->get<Mesh<2>>();

          const Connectivity<2>& connectivity = mesh.connectivity();
          REQUIRE(
            is_valid.template
            operator()<ItemType::node,
                       ItemType::node>(connectivity,
                                       StencilManager::instance()
                                         .getNodeToCellStencilArray(connectivity,
                                                                    StencilDescriptor{1, StencilDescriptor::
                                                                                           ConnectionType::by_nodes}),
                                       1));

          REQUIRE(
            is_valid.template
            operator()<ItemType::node,
                       ItemType::edge>(connectivity,
                                       StencilManager::instance()
                                         .getNodeToCellStencilArray(connectivity,
                                                                    StencilDescriptor{1, StencilDescriptor::
                                                                                           ConnectionType::by_edges}),
                                       1));

          REQUIRE(
            is_valid.template
            operator()<ItemType::node,
                       ItemType::face>(connectivity,
                                       StencilManager::instance()
                                         .getNodeToCellStencilArray(connectivity,
                                                                    StencilDescriptor{1, StencilDescriptor::
                                                                                           ConnectionType::by_faces}),
                                       1));
        }
      }

      SECTION("3D")
      {
        SECTION("carteian")
        {
          const auto& mesh = *MeshDataBaseForTests::get().cartesian3DMesh()->get<Mesh<3>>();

          const Connectivity<3>& connectivity = mesh.connectivity();
          REQUIRE(
            is_valid.template
            operator()<ItemType::node,
                       ItemType::node>(connectivity,
                                       StencilManager::instance()
                                         .getNodeToCellStencilArray(connectivity,
                                                                    StencilDescriptor{1, StencilDescriptor::
                                                                                           ConnectionType::by_nodes}),
                                       1));

          REQUIRE(
            is_valid.template
            operator()<ItemType::node,
                       ItemType::edge>(connectivity,
                                       StencilManager::instance()
                                         .getNodeToCellStencilArray(connectivity,
                                                                    StencilDescriptor{1, StencilDescriptor::
                                                                                           ConnectionType::by_edges}),
                                       1));

          REQUIRE(
            is_valid.template
            operator()<ItemType::node,
                       ItemType::face>(connectivity,
                                       StencilManager::instance()
                                         .getNodeToCellStencilArray(connectivity,
                                                                    StencilDescriptor{1, StencilDescriptor::
                                                                                           ConnectionType::by_faces}),
                                       1));
        }

        SECTION("hybrid")
        {
          const auto& mesh = *MeshDataBaseForTests::get().hybrid3DMesh()->get<Mesh<3>>();

          const Connectivity<3>& connectivity = mesh.connectivity();
          REQUIRE(
            is_valid.template
            operator()<ItemType::node,
                       ItemType::node>(connectivity,
                                       StencilManager::instance()
                                         .getNodeToCellStencilArray(connectivity,
                                                                    StencilDescriptor{1, StencilDescriptor::
                                                                                           ConnectionType::by_nodes}),
                                       1));

          REQUIRE(
            is_valid.template
            operator()<ItemType::node,
                       ItemType::edge>(connectivity,
                                       StencilManager::instance()
                                         .getNodeToCellStencilArray(connectivity,
                                                                    StencilDescriptor{1, StencilDescriptor::
                                                                                           ConnectionType::by_edges}),
                                       1));

          REQUIRE(
            is_valid.template
            operator()<ItemType::node,
                       ItemType::face>(connectivity,
                                       StencilManager::instance()
                                         .getNodeToCellStencilArray(connectivity,
                                                                    StencilDescriptor{1, StencilDescriptor::
                                                                                           ConnectionType::by_faces}),
                                       1));
        }
      }
    }

    SECTION("2 layers")
    {
      NbGhostLayersTester nb_ghost_layers_tester(2);

      SECTION("1D")
      {
        SECTION("cartesian")
        {
          auto mesh_v = CartesianMeshBuilder(TinyVector<1>{0}, TinyVector<1>{1}, TinyVector<1, uint64_t>(20)).mesh();

          const auto& mesh = *mesh_v->get<Mesh<1>>();

          REQUIRE(
            is_valid.template
            operator()<ItemType::node,
                       ItemType::node>(mesh.connectivity(),
                                       StencilManager::instance()
                                         .getNodeToCellStencilArray(mesh.connectivity(),
                                                                    StencilDescriptor{2, StencilDescriptor::
                                                                                           ConnectionType::by_nodes}),
                                       2));

          REQUIRE(
            is_valid.template
            operator()<ItemType::node,
                       ItemType::edge>(mesh.connectivity(),
                                       StencilManager::instance()
                                         .getNodeToCellStencilArray(mesh.connectivity(),
                                                                    StencilDescriptor{2, StencilDescriptor::
                                                                                           ConnectionType::by_edges}),
                                       2));

          REQUIRE(
            is_valid.template
            operator()<ItemType::node,
                       ItemType::face>(mesh.connectivity(),
                                       StencilManager::instance()
                                         .getNodeToCellStencilArray(mesh.connectivity(),
                                                                    StencilDescriptor{2, StencilDescriptor::
                                                                                           ConnectionType::by_faces}),
                                       2));
        }
      }

      SECTION("2D")
      {
        SECTION("cartesian")
        {
          auto mesh_v =
            CartesianMeshBuilder(TinyVector<2>{0, 0}, TinyVector<2>{1, 2}, TinyVector<2, uint64_t>(5, 7)).mesh();
          const auto& mesh = *mesh_v->get<Mesh<2>>();

          REQUIRE(
            is_valid.template
            operator()<ItemType::node,
                       ItemType::node>(mesh.connectivity(),
                                       StencilManager::instance()
                                         .getNodeToCellStencilArray(mesh.connectivity(),
                                                                    StencilDescriptor{2, StencilDescriptor::
                                                                                           ConnectionType::by_nodes}),
                                       2));

          REQUIRE(
            is_valid.template
            operator()<ItemType::node,
                       ItemType::edge>(mesh.connectivity(),
                                       StencilManager::instance()
                                         .getNodeToCellStencilArray(mesh.connectivity(),
                                                                    StencilDescriptor{2, StencilDescriptor::
                                                                                           ConnectionType::by_edges}),
                                       2));

          REQUIRE(
            is_valid.template
            operator()<ItemType::node,
                       ItemType::face>(mesh.connectivity(),
                                       StencilManager::instance()
                                         .getNodeToCellStencilArray(mesh.connectivity(),
                                                                    StencilDescriptor{2, StencilDescriptor::
                                                                                           ConnectionType::by_faces}),
                                       2));
        }
      }

      SECTION("3D")
      {
        SECTION("carteian")
        {
          auto mesh_v =
            CartesianMeshBuilder(TinyVector<3>{0, 0, 0}, TinyVector<3>{1, 1, 2}, TinyVector<3, uint64_t>(3, 4, 5))
              .mesh();
          const auto& mesh = *mesh_v->get<Mesh<3>>();

          REQUIRE(
            is_valid.template
            operator()<ItemType::node,
                       ItemType::node>(mesh.connectivity(),
                                       StencilManager::instance()
                                         .getNodeToCellStencilArray(mesh.connectivity(),
                                                                    StencilDescriptor{2, StencilDescriptor::
                                                                                           ConnectionType::by_nodes}),
                                       2));

          REQUIRE(
            is_valid.template
            operator()<ItemType::node,
                       ItemType::edge>(mesh.connectivity(),
                                       StencilManager::instance()
                                         .getNodeToCellStencilArray(mesh.connectivity(),
                                                                    StencilDescriptor{2, StencilDescriptor::
                                                                                           ConnectionType::by_edges}),
                                       2));

          REQUIRE(
            is_valid.template
            operator()<ItemType::node,
                       ItemType::face>(mesh.connectivity(),
                                       StencilManager::instance()
                                         .getNodeToCellStencilArray(mesh.connectivity(),
                                                                    StencilDescriptor{2, StencilDescriptor::
                                                                                           ConnectionType::by_faces}),
                                       2));
        }
      }
    }
  }

  SECTION("Stencil using symmetries")
  {
    auto check_ghost_nodes_have_empty_stencils = [](const auto& stencil_array, const auto& connecticity) {
      bool is_empty = true;

      auto node_is_owned = connecticity.nodeIsOwned();

      for (NodeId node_id = 0; node_id < connecticity.numberOfNodes(); ++node_id) {
        if (not node_is_owned[node_id]) {
          is_empty &= (stencil_array[node_id].size() == 0);
          for (size_t i_symmetry_stencil = 0;
               i_symmetry_stencil < stencil_array.symmetryBoundaryStencilArrayList().size(); ++i_symmetry_stencil) {
            is_empty &=
              (stencil_array.symmetryBoundaryStencilArrayList()[i_symmetry_stencil].stencilArray()[node_id].size() ==
               0);
          }
        }
      }

      return is_empty;
    };

    auto are_symmetry_stencils_valid =
      []<ItemType source_item_type, ItemType connecting_item_type>(const auto& stencil_array, const auto& mesh,
                                                                   const size_t number_of_layers) {
        bool are_valid_symmetries = true;

        constexpr const size_t Dimension = std::decay_t<decltype(mesh)>::Dimension;

        auto connecting_to_cell_matrix =
          mesh.connectivity().template getItemToItemMatrix<connecting_item_type, ItemType::cell>();
        auto cell_to_connecting_item_matrix =
          mesh.connectivity().template getItemToItemMatrix<ItemType::cell, connecting_item_type>();
        //        auto source_item_is_owned        = mesh.connectivity().cellIsOwned();
        auto cell_number          = mesh.connectivity().cellNumber();
        auto source_item_is_owned = mesh.connectivity().template isOwned<source_item_type>();

        auto connecting_number = mesh.connectivity().template number<connecting_item_type>();

        auto source_item_to_connecting_item_matrix =
          [](const auto& given_connectivity) -> ItemToItemMatrix<source_item_type, connecting_item_type> {
          if constexpr (ItemTypeId<Dimension>::itemTId(source_item_type) ==
                        ItemTypeId<Dimension>::itemTId(connecting_item_type)) {
            return ConnectivityMatrix{};
          } else {
            return given_connectivity.template getItemToItemMatrix<source_item_type, connecting_item_type>();
          }
        }(mesh.connectivity());

        using SourceItemId     = ItemIdT<source_item_type>;
        using ConnectingItemId = ItemIdT<connecting_item_type>;
        using MeshType         = std::decay_t<decltype(mesh)>;

        for (size_t i_symmetry_stencil = 0;
             i_symmetry_stencil < stencil_array.symmetryBoundaryStencilArrayList().size(); ++i_symmetry_stencil) {
          const IBoundaryDescriptor& boundary_descriptor =
            stencil_array.symmetryBoundaryStencilArrayList()[i_symmetry_stencil].boundaryDescriptor();

          auto boundary_stencil   = stencil_array.symmetryBoundaryStencilArrayList()[i_symmetry_stencil].stencilArray();
          auto boundary_face_list = getMeshFaceBoundary(mesh, boundary_descriptor);

          std::set<ConnectingItemId> sym_connecting_item_set;

          if constexpr (ItemTypeId<MeshType::Dimension>::itemTId(connecting_item_type) ==
                        ItemTypeId<MeshType::Dimension>::itemTId(ItemType::face)) {
            for (size_t i_face = 0; i_face < boundary_face_list.faceList().size(); ++i_face) {
              const FaceId face_id = boundary_face_list.faceList()[i_face];
              sym_connecting_item_set.insert(ConnectingItemId{FaceId::base_type{face_id}});
            }

          } else {
            auto face_to_connecting_item_matrix =
              mesh.connectivity().template getItemToItemMatrix<ItemType::face, connecting_item_type>();

            for (size_t i_face = 0; i_face < boundary_face_list.faceList().size(); ++i_face) {
              const FaceId face_id      = boundary_face_list.faceList()[i_face];
              auto connecting_item_list = face_to_connecting_item_matrix[face_id];
              for (size_t i_connecting_item = 0; i_connecting_item < connecting_item_list.size(); ++i_connecting_item) {
                sym_connecting_item_set.insert(connecting_item_list[i_connecting_item]);
              }
            }
          }

          for (SourceItemId source_item_id = 0; source_item_id < mesh.template numberOf<source_item_type>();
               ++source_item_id) {
            if (not source_item_is_owned[source_item_id]) {
              are_valid_symmetries &= (boundary_stencil[source_item_id].size() == 0);
            } else {
              std::vector<CellId> expected_stencil;

              std::set<CellId> marked_cell_set;
              if constexpr (source_item_type == ItemType::cell) {
                marked_cell_set.insert(source_item_id);
              }

              std::set<ConnectingItemId> marked_connecting_item_set;
              std::set<ConnectingItemId> layer_connecting_item_set;
              if constexpr (ItemTypeId<Dimension>::itemTId(source_item_type) ==
                            ItemTypeId<Dimension>::itemTId(connecting_item_type)) {
                const ConnectingItemId connecting_item_id =
                  static_cast<typename SourceItemId::base_type>(source_item_id);
                layer_connecting_item_set.insert(connecting_item_id);
                marked_connecting_item_set.insert(connecting_item_id);
              } else {
                auto source_item_to_connecting_item_list = source_item_to_connecting_item_matrix[source_item_id];
                for (size_t i_connecting_item = 0; i_connecting_item < source_item_to_connecting_item_list.size();
                     ++i_connecting_item) {
                  const ConnectingItemId connecting_item_id = source_item_to_connecting_item_list[i_connecting_item];
                  layer_connecting_item_set.insert(connecting_item_id);
                  marked_connecting_item_set.insert(connecting_item_id);
                }
              }

              std::set<ConnectingItemId> marked_sym_connecting_item_set;
              std::set<ConnectingItemId> layer_sym_connecting_item_set;

              for (auto&& connecting_item_id : marked_connecting_item_set) {
                if (sym_connecting_item_set.contains(connecting_item_id)) {
                  marked_sym_connecting_item_set.insert(connecting_item_id);
                  layer_sym_connecting_item_set.insert(connecting_item_id);
                }
              }

              std::set<CellId> marked_sym_cell_set;

              for (size_t i_layer = 0; i_layer < number_of_layers; ++i_layer) {
                std::set<CellId, std::function<bool(CellId, CellId)>> cell_set(
                  [=](CellId cell_0, CellId cell_1) { return cell_number[cell_0] < cell_number[cell_1]; });

                for (auto&& connecting_item_id : layer_connecting_item_set) {
                  auto connecting_item_to_cell_list = connecting_to_cell_matrix[connecting_item_id];
                  for (size_t i_connecting_item_of_cell = 0;
                       i_connecting_item_of_cell < connecting_item_to_cell_list.size(); ++i_connecting_item_of_cell) {
                    const CellId connecting_item_id_of_cell = connecting_item_to_cell_list[i_connecting_item_of_cell];
                    if (not marked_cell_set.contains(connecting_item_id_of_cell)) {
                      cell_set.insert(connecting_item_id_of_cell);
                      marked_cell_set.insert(connecting_item_id_of_cell);
                    }
                  }
                }

                std::set<CellId, std::function<bool(CellId, CellId)>> sym_cell_set(
                  [=](CellId cell_0, CellId cell_1) { return cell_number[cell_0] < cell_number[cell_1]; });

                for (auto&& connecting_item_id : layer_sym_connecting_item_set) {
                  auto connecting_item_to_cell_list = connecting_to_cell_matrix[connecting_item_id];
                  for (size_t i_connecting_item_of_cell = 0;
                       i_connecting_item_of_cell < connecting_item_to_cell_list.size(); ++i_connecting_item_of_cell) {
                    const CellId connecting_item_id_of_cell = connecting_item_to_cell_list[i_connecting_item_of_cell];
                    if (not marked_sym_cell_set.contains(connecting_item_id_of_cell)) {
                      sym_cell_set.insert(connecting_item_id_of_cell);
                      marked_sym_cell_set.insert(connecting_item_id_of_cell);
                    }
                  }
                }

                for (auto&& set_sym_cell_id : sym_cell_set) {
                  expected_stencil.push_back(set_sym_cell_id);
                }

                layer_connecting_item_set.clear();
                for (auto&& layer_cell_id : cell_set) {
                  auto cell_to_connecting_item_list = cell_to_connecting_item_matrix[layer_cell_id];
                  for (size_t i_connecting_item = 0; i_connecting_item < cell_to_connecting_item_list.size();
                       ++i_connecting_item) {
                    const ConnectingItemId connecting_item_id = cell_to_connecting_item_list[i_connecting_item];
                    if (not marked_connecting_item_set.contains(connecting_item_id)) {
                      layer_connecting_item_set.insert(connecting_item_id);
                      marked_connecting_item_set.insert(connecting_item_id);
                    }
                  }
                }

                layer_sym_connecting_item_set.clear();

                for (auto&& connecting_item_id : layer_connecting_item_set) {
                  if (sym_connecting_item_set.contains(connecting_item_id)) {
                    if (not marked_sym_connecting_item_set.contains(connecting_item_id)) {
                      marked_sym_connecting_item_set.insert(connecting_item_id);
                      layer_sym_connecting_item_set.insert(connecting_item_id);
                    }
                  }
                }

                for (auto layer_sym_cell_id : sym_cell_set) {
                  auto cell_to_connecting_item_list = cell_to_connecting_item_matrix[layer_sym_cell_id];
                  for (size_t i_connecting_item = 0; i_connecting_item < cell_to_connecting_item_list.size();
                       ++i_connecting_item) {
                    const ConnectingItemId connecting_item_id = cell_to_connecting_item_list[i_connecting_item];
                    if (not marked_sym_connecting_item_set.contains(connecting_item_id)) {
                      marked_sym_connecting_item_set.insert(connecting_item_id);
                      layer_sym_connecting_item_set.insert(connecting_item_id);
                    }
                  }
                }
              }

              auto cell_stencil = boundary_stencil[source_item_id];

              if (cell_stencil.size() != expected_stencil.size()) {
                are_valid_symmetries = false;
              }

              auto i_set_cell = expected_stencil.begin();
              for (size_t index = 0; index < cell_stencil.size(); ++index, ++i_set_cell) {
                if (*i_set_cell != cell_stencil[index]) {
                  are_valid_symmetries = false;
                }
              }
            }
          }
        }

        return are_valid_symmetries;
      };

    SECTION("1 layer")
    {
      SECTION("1D")
      {
        StencilManager::BoundaryDescriptorList list;
        list.emplace_back(std::make_shared<NamedBoundaryDescriptor>("XMIN"));
        list.emplace_back(std::make_shared<NamedBoundaryDescriptor>("XMAX"));

        SECTION("cartesian")
        {
          const auto& mesh = *MeshDataBaseForTests::get().cartesian1DMesh()->get<Mesh<1>>();

          const Connectivity<1>& connectivity = mesh.connectivity();

          {
            auto stencil_array =
              StencilManager::instance()
                .getNodeToCellStencilArray(connectivity,
                                           StencilDescriptor{1, StencilDescriptor::ConnectionType::by_nodes}, list);

            REQUIRE(stencil_array.symmetryBoundaryStencilArrayList().size() == 2);
            REQUIRE(check_ghost_nodes_have_empty_stencils(stencil_array, connectivity));
            REQUIRE(
              are_symmetry_stencils_valid.template operator()<ItemType::node, ItemType::node>(stencil_array, mesh, 1));
            REQUIRE(is_valid.template operator()<ItemType::node, ItemType::node>(connectivity, stencil_array, 1));
          }

          {
            auto stencil_array =
              StencilManager::instance()
                .getNodeToCellStencilArray(connectivity,
                                           StencilDescriptor{1, StencilDescriptor::ConnectionType::by_edges}, list);

            REQUIRE(stencil_array.symmetryBoundaryStencilArrayList().size() == 2);
            REQUIRE(check_ghost_nodes_have_empty_stencils(stencil_array, connectivity));
            REQUIRE(
              are_symmetry_stencils_valid.template operator()<ItemType::node, ItemType::edge>(stencil_array, mesh, 1));
            REQUIRE(is_valid.template operator()<ItemType::node, ItemType::node>(connectivity, stencil_array, 1));
          }

          {
            auto stencil_array =
              StencilManager::instance()
                .getNodeToCellStencilArray(connectivity,
                                           StencilDescriptor{1, StencilDescriptor::ConnectionType::by_faces}, list);

            REQUIRE(stencil_array.symmetryBoundaryStencilArrayList().size() == 2);
            REQUIRE(check_ghost_nodes_have_empty_stencils(stencil_array, connectivity));
            REQUIRE(
              are_symmetry_stencils_valid.template operator()<ItemType::node, ItemType::face>(stencil_array, mesh, 1));
            REQUIRE(is_valid.template operator()<ItemType::node, ItemType::node>(connectivity, stencil_array, 1));
          }
        }

        SECTION("unordered")
        {
          const auto& mesh = *MeshDataBaseForTests::get().unordered1DMesh()->get<Mesh<1>>();

          const Connectivity<1>& connectivity = mesh.connectivity();

          {
            auto stencil_array =
              StencilManager::instance()
                .getNodeToCellStencilArray(connectivity,
                                           StencilDescriptor{1, StencilDescriptor::ConnectionType::by_nodes}, list);

            REQUIRE(stencil_array.symmetryBoundaryStencilArrayList().size() == 2);
            REQUIRE(check_ghost_nodes_have_empty_stencils(stencil_array, connectivity));
            REQUIRE(
              are_symmetry_stencils_valid.template operator()<ItemType::node, ItemType::node>(stencil_array, mesh, 1));
            REQUIRE(is_valid.template operator()<ItemType::node, ItemType::node>(connectivity, stencil_array, 1));
          }

          {
            auto stencil_array =
              StencilManager::instance()
                .getNodeToCellStencilArray(connectivity,
                                           StencilDescriptor{1, StencilDescriptor::ConnectionType::by_edges}, list);

            REQUIRE(stencil_array.symmetryBoundaryStencilArrayList().size() == 2);
            REQUIRE(check_ghost_nodes_have_empty_stencils(stencil_array, connectivity));
            REQUIRE(
              are_symmetry_stencils_valid.template operator()<ItemType::node, ItemType::edge>(stencil_array, mesh, 1));
            REQUIRE(is_valid.template operator()<ItemType::node, ItemType::node>(connectivity, stencil_array, 1));
          }

          {
            auto stencil_array =
              StencilManager::instance()
                .getNodeToCellStencilArray(connectivity,
                                           StencilDescriptor{1, StencilDescriptor::ConnectionType::by_faces}, list);

            REQUIRE(stencil_array.symmetryBoundaryStencilArrayList().size() == 2);
            REQUIRE(check_ghost_nodes_have_empty_stencils(stencil_array, connectivity));
            REQUIRE(
              are_symmetry_stencils_valid.template operator()<ItemType::node, ItemType::face>(stencil_array, mesh, 1));
            REQUIRE(is_valid.template operator()<ItemType::node, ItemType::node>(connectivity, stencil_array, 1));
          }
        }
      }

      SECTION("2D")
      {
        StencilManager::BoundaryDescriptorList list;
        list.emplace_back(std::make_shared<NamedBoundaryDescriptor>("XMIN"));
        list.emplace_back(std::make_shared<NamedBoundaryDescriptor>("XMAX"));
        list.emplace_back(std::make_shared<NamedBoundaryDescriptor>("YMIN"));
        list.emplace_back(std::make_shared<NamedBoundaryDescriptor>("YMAX"));

        SECTION("cartesian")
        {
          const auto& mesh = *MeshDataBaseForTests::get().cartesian2DMesh()->get<Mesh<2>>();

          const Connectivity<2>& connectivity = mesh.connectivity();

          {
            auto stencil_array =
              StencilManager::instance()
                .getNodeToCellStencilArray(connectivity,
                                           StencilDescriptor{1, StencilDescriptor::ConnectionType::by_nodes}, list);

            REQUIRE(stencil_array.symmetryBoundaryStencilArrayList().size() == 4);
            REQUIRE(check_ghost_nodes_have_empty_stencils(stencil_array, connectivity));
            REQUIRE(
              are_symmetry_stencils_valid.template operator()<ItemType::node, ItemType::node>(stencil_array, mesh, 1));
            REQUIRE(is_valid.template operator()<ItemType::node, ItemType::node>(connectivity, stencil_array, 1));
          }

          {
            auto stencil_array =
              StencilManager::instance()
                .getNodeToCellStencilArray(connectivity,
                                           StencilDescriptor{1, StencilDescriptor::ConnectionType::by_edges}, list);

            REQUIRE(stencil_array.symmetryBoundaryStencilArrayList().size() == 4);
            REQUIRE(check_ghost_nodes_have_empty_stencils(stencil_array, connectivity));
            REQUIRE(
              are_symmetry_stencils_valid.template operator()<ItemType::node, ItemType::edge>(stencil_array, mesh, 1));
            REQUIRE(is_valid.template operator()<ItemType::node, ItemType::edge>(connectivity, stencil_array, 1));
          }

          {
            auto stencil_array =
              StencilManager::instance()
                .getNodeToCellStencilArray(connectivity,
                                           StencilDescriptor{1, StencilDescriptor::ConnectionType::by_faces}, list);

            REQUIRE(stencil_array.symmetryBoundaryStencilArrayList().size() == 4);
            REQUIRE(check_ghost_nodes_have_empty_stencils(stencil_array, connectivity));
            REQUIRE(
              are_symmetry_stencils_valid.template operator()<ItemType::node, ItemType::face>(stencil_array, mesh, 1));
            REQUIRE(is_valid.template operator()<ItemType::node, ItemType::face>(connectivity, stencil_array, 1));
          }
        }

        SECTION("hybrid")
        {
          const auto& mesh = *MeshDataBaseForTests::get().hybrid2DMesh()->get<Mesh<2>>();

          const Connectivity<2>& connectivity = mesh.connectivity();

          {
            auto stencil_array =
              StencilManager::instance()
                .getNodeToCellStencilArray(connectivity,
                                           StencilDescriptor{1, StencilDescriptor::ConnectionType::by_nodes}, list);

            REQUIRE(stencil_array.symmetryBoundaryStencilArrayList().size() == 4);
            REQUIRE(check_ghost_nodes_have_empty_stencils(stencil_array, connectivity));
            REQUIRE(
              are_symmetry_stencils_valid.template operator()<ItemType::node, ItemType::node>(stencil_array, mesh, 1));
            REQUIRE(is_valid.template operator()<ItemType::node, ItemType::node>(connectivity, stencil_array, 1));
          }

          {
            auto stencil_array =
              StencilManager::instance()
                .getNodeToCellStencilArray(connectivity,
                                           StencilDescriptor{1, StencilDescriptor::ConnectionType::by_edges}, list);

            REQUIRE(stencil_array.symmetryBoundaryStencilArrayList().size() == 4);
            REQUIRE(check_ghost_nodes_have_empty_stencils(stencil_array, connectivity));
            REQUIRE(
              are_symmetry_stencils_valid.template operator()<ItemType::node, ItemType::edge>(stencil_array, mesh, 1));
            REQUIRE(is_valid.template operator()<ItemType::node, ItemType::edge>(connectivity, stencil_array, 1));
          }

          {
            auto stencil_array =
              StencilManager::instance()
                .getNodeToCellStencilArray(connectivity,
                                           StencilDescriptor{1, StencilDescriptor::ConnectionType::by_faces}, list);

            REQUIRE(stencil_array.symmetryBoundaryStencilArrayList().size() == 4);
            REQUIRE(check_ghost_nodes_have_empty_stencils(stencil_array, connectivity));
            REQUIRE(
              are_symmetry_stencils_valid.template operator()<ItemType::node, ItemType::face>(stencil_array, mesh, 1));
            REQUIRE(is_valid.template operator()<ItemType::node, ItemType::face>(connectivity, stencil_array, 1));
          }
        }
      }

      SECTION("3D")
      {
        StencilManager::BoundaryDescriptorList list;
        list.emplace_back(std::make_shared<NamedBoundaryDescriptor>("XMIN"));
        list.emplace_back(std::make_shared<NamedBoundaryDescriptor>("XMAX"));
        list.emplace_back(std::make_shared<NamedBoundaryDescriptor>("YMIN"));
        list.emplace_back(std::make_shared<NamedBoundaryDescriptor>("YMAX"));
        list.emplace_back(std::make_shared<NamedBoundaryDescriptor>("ZMIN"));
        list.emplace_back(std::make_shared<NamedBoundaryDescriptor>("ZMAX"));

        SECTION("cartesian")
        {
          const auto& mesh = *MeshDataBaseForTests::get().cartesian3DMesh()->get<Mesh<3>>();

          const Connectivity<3>& connectivity = mesh.connectivity();
          {
            auto stencil_array =
              StencilManager::instance()
                .getNodeToCellStencilArray(connectivity,
                                           StencilDescriptor{1, StencilDescriptor::ConnectionType::by_nodes}, list);

            REQUIRE(stencil_array.symmetryBoundaryStencilArrayList().size() == 6);
            REQUIRE(check_ghost_nodes_have_empty_stencils(stencil_array, connectivity));
            REQUIRE(
              are_symmetry_stencils_valid.template operator()<ItemType::node, ItemType::node>(stencil_array, mesh, 1));
            REQUIRE(is_valid.template operator()<ItemType::node, ItemType::node>(connectivity, stencil_array, 1));
          }

          {
            auto stencil_array =
              StencilManager::instance()
                .getNodeToCellStencilArray(connectivity,
                                           StencilDescriptor{1, StencilDescriptor::ConnectionType::by_edges}, list);

            REQUIRE(stencil_array.symmetryBoundaryStencilArrayList().size() == 6);
            REQUIRE(check_ghost_nodes_have_empty_stencils(stencil_array, connectivity));
            REQUIRE(
              are_symmetry_stencils_valid.template operator()<ItemType::node, ItemType::edge>(stencil_array, mesh, 1));
            REQUIRE(is_valid.template operator()<ItemType::node, ItemType::edge>(connectivity, stencil_array, 1));
          }

          {
            auto stencil_array =
              StencilManager::instance()
                .getNodeToCellStencilArray(connectivity,
                                           StencilDescriptor{1, StencilDescriptor::ConnectionType::by_faces}, list);

            REQUIRE(stencil_array.symmetryBoundaryStencilArrayList().size() == 6);
            REQUIRE(check_ghost_nodes_have_empty_stencils(stencil_array, connectivity));
            REQUIRE(
              are_symmetry_stencils_valid.template operator()<ItemType::node, ItemType::face>(stencil_array, mesh, 1));
            REQUIRE(is_valid.template operator()<ItemType::node, ItemType::face>(connectivity, stencil_array, 1));
          }
        }

        SECTION("hybrid")
        {
          const auto& mesh = *MeshDataBaseForTests::get().hybrid3DMesh()->get<Mesh<3>>();

          const Connectivity<3>& connectivity = mesh.connectivity();
          {
            auto stencil_array =
              StencilManager::instance()
                .getNodeToCellStencilArray(connectivity,
                                           StencilDescriptor{1, StencilDescriptor::ConnectionType::by_nodes}, list);

            REQUIRE(stencil_array.symmetryBoundaryStencilArrayList().size() == 6);
            REQUIRE(check_ghost_nodes_have_empty_stencils(stencil_array, connectivity));
            REQUIRE(
              are_symmetry_stencils_valid.template operator()<ItemType::node, ItemType::node>(stencil_array, mesh, 1));
            REQUIRE(is_valid.template operator()<ItemType::node, ItemType::node>(connectivity, stencil_array, 1));
          }

          {
            auto stencil_array =
              StencilManager::instance()
                .getNodeToCellStencilArray(connectivity,
                                           StencilDescriptor{1, StencilDescriptor::ConnectionType::by_edges}, list);

            REQUIRE(stencil_array.symmetryBoundaryStencilArrayList().size() == 6);
            REQUIRE(check_ghost_nodes_have_empty_stencils(stencil_array, connectivity));
            REQUIRE(
              are_symmetry_stencils_valid.template operator()<ItemType::node, ItemType::edge>(stencil_array, mesh, 1));
            REQUIRE(is_valid.template operator()<ItemType::node, ItemType::edge>(connectivity, stencil_array, 1));
          }

          {
            auto stencil_array =
              StencilManager::instance()
                .getNodeToCellStencilArray(connectivity,
                                           StencilDescriptor{1, StencilDescriptor::ConnectionType::by_faces}, list);

            REQUIRE(stencil_array.symmetryBoundaryStencilArrayList().size() == 6);
            REQUIRE(check_ghost_nodes_have_empty_stencils(stencil_array, connectivity));
            REQUIRE(
              are_symmetry_stencils_valid.template operator()<ItemType::node, ItemType::face>(stencil_array, mesh, 1));
            REQUIRE(is_valid.template operator()<ItemType::node, ItemType::face>(connectivity, stencil_array, 1));
          }
        }
      }
    }

    SECTION("2 layers")
    {
      NbGhostLayersTester nb_ghost_layers_tester(2);

      SECTION("1D")
      {
        StencilManager::BoundaryDescriptorList list;
        list.emplace_back(std::make_shared<NamedBoundaryDescriptor>("XMIN"));
        list.emplace_back(std::make_shared<NamedBoundaryDescriptor>("XMAX"));

        SECTION("cartesian")
        {
          auto mesh_v = CartesianMeshBuilder(TinyVector<1>{0}, TinyVector<1>{1}, TinyVector<1, uint64_t>(20)).mesh();
          const auto& mesh = *mesh_v->get<Mesh<1>>();

          const Connectivity<1>& connectivity = mesh.connectivity();

          {
            auto stencil_array =
              StencilManager::instance()
                .getNodeToCellStencilArray(connectivity,
                                           StencilDescriptor{2, StencilDescriptor::ConnectionType::by_nodes}, list);

            REQUIRE(stencil_array.symmetryBoundaryStencilArrayList().size() == 2);
            REQUIRE(check_ghost_nodes_have_empty_stencils(stencil_array, connectivity));
            REQUIRE(
              are_symmetry_stencils_valid.template operator()<ItemType::node, ItemType::node>(stencil_array, mesh, 2));
            REQUIRE(is_valid.template operator()<ItemType::node, ItemType::node>(connectivity, stencil_array, 2));
          }

          {
            auto stencil_array =
              StencilManager::instance()
                .getNodeToCellStencilArray(connectivity,
                                           StencilDescriptor{2, StencilDescriptor::ConnectionType::by_edges}, list);

            REQUIRE(stencil_array.symmetryBoundaryStencilArrayList().size() == 2);
            REQUIRE(check_ghost_nodes_have_empty_stencils(stencil_array, connectivity));
            REQUIRE(
              are_symmetry_stencils_valid.template operator()<ItemType::node, ItemType::edge>(stencil_array, mesh, 2));
            REQUIRE(is_valid.template operator()<ItemType::node, ItemType::node>(connectivity, stencil_array, 2));
          }

          {
            auto stencil_array =
              StencilManager::instance()
                .getNodeToCellStencilArray(connectivity,
                                           StencilDescriptor{2, StencilDescriptor::ConnectionType::by_faces}, list);

            REQUIRE(stencil_array.symmetryBoundaryStencilArrayList().size() == 2);
            REQUIRE(check_ghost_nodes_have_empty_stencils(stencil_array, connectivity));
            REQUIRE(
              are_symmetry_stencils_valid.template operator()<ItemType::node, ItemType::face>(stencil_array, mesh, 2));
            REQUIRE(is_valid.template operator()<ItemType::node, ItemType::node>(connectivity, stencil_array, 2));
          }
        }
      }

      SECTION("2D")
      {
        StencilManager::BoundaryDescriptorList list;
        list.emplace_back(std::make_shared<NamedBoundaryDescriptor>("XMIN"));
        list.emplace_back(std::make_shared<NamedBoundaryDescriptor>("XMAX"));
        list.emplace_back(std::make_shared<NamedBoundaryDescriptor>("YMIN"));
        list.emplace_back(std::make_shared<NamedBoundaryDescriptor>("YMAX"));

        SECTION("cartesian")
        {
          auto mesh_v =
            CartesianMeshBuilder(TinyVector<2>{0, 0}, TinyVector<2>{1, 2}, TinyVector<2, uint64_t>(5, 7)).mesh();
          const auto& mesh = *mesh_v->get<Mesh<2>>();

          const Connectivity<2>& connectivity = mesh.connectivity();
          {
            auto stencil_array =
              StencilManager::instance()
                .getNodeToCellStencilArray(connectivity,
                                           StencilDescriptor{2, StencilDescriptor::ConnectionType::by_nodes}, list);

            REQUIRE(stencil_array.symmetryBoundaryStencilArrayList().size() == 4);
            REQUIRE(check_ghost_nodes_have_empty_stencils(stencil_array, connectivity));
            REQUIRE(
              are_symmetry_stencils_valid.template operator()<ItemType::node, ItemType::node>(stencil_array, mesh, 2));
            REQUIRE(is_valid.template operator()<ItemType::node, ItemType::node>(connectivity, stencil_array, 2));
          }

          {
            auto stencil_array =
              StencilManager::instance()
                .getNodeToCellStencilArray(connectivity,
                                           StencilDescriptor{2, StencilDescriptor::ConnectionType::by_edges}, list);

            REQUIRE(stencil_array.symmetryBoundaryStencilArrayList().size() == 4);
            REQUIRE(check_ghost_nodes_have_empty_stencils(stencil_array, connectivity));
            REQUIRE(
              are_symmetry_stencils_valid.template operator()<ItemType::node, ItemType::edge>(stencil_array, mesh, 2));
            REQUIRE(is_valid.template operator()<ItemType::node, ItemType::edge>(connectivity, stencil_array, 2));
          }

          {
            auto stencil_array =
              StencilManager::instance()
                .getNodeToCellStencilArray(connectivity,
                                           StencilDescriptor{2, StencilDescriptor::ConnectionType::by_faces}, list);

            REQUIRE(stencil_array.symmetryBoundaryStencilArrayList().size() == 4);
            REQUIRE(check_ghost_nodes_have_empty_stencils(stencil_array, connectivity));
            REQUIRE(
              are_symmetry_stencils_valid.template operator()<ItemType::node, ItemType::face>(stencil_array, mesh, 2));
            REQUIRE(not(is_valid.template operator()<ItemType::node, ItemType::node>(connectivity, stencil_array, 2)));
            REQUIRE(is_valid.template operator()<ItemType::node, ItemType::face>(connectivity, stencil_array, 2));
          }
        }
      }

      SECTION("3D")
      {
        StencilManager::BoundaryDescriptorList list;
        list.emplace_back(std::make_shared<NamedBoundaryDescriptor>("XMIN"));
        list.emplace_back(std::make_shared<NamedBoundaryDescriptor>("XMAX"));
        list.emplace_back(std::make_shared<NamedBoundaryDescriptor>("YMIN"));
        list.emplace_back(std::make_shared<NamedBoundaryDescriptor>("YMAX"));
        list.emplace_back(std::make_shared<NamedBoundaryDescriptor>("ZMIN"));
        list.emplace_back(std::make_shared<NamedBoundaryDescriptor>("ZMAX"));

        SECTION("cartesian")
        {
          auto mesh_v =
            CartesianMeshBuilder(TinyVector<3>{0, 0, 0}, TinyVector<3>{1, 1, 2}, TinyVector<3, uint64_t>(3, 4, 5))
              .mesh();
          const auto& mesh = *mesh_v->get<Mesh<3>>();

          const Connectivity<3>& connectivity = mesh.connectivity();
          {
            auto stencil_array =
              StencilManager::instance()
                .getNodeToCellStencilArray(connectivity,
                                           StencilDescriptor{2, StencilDescriptor::ConnectionType::by_nodes}, list);

            REQUIRE(stencil_array.symmetryBoundaryStencilArrayList().size() == 6);
            REQUIRE(check_ghost_nodes_have_empty_stencils(stencil_array, connectivity));
            REQUIRE(
              are_symmetry_stencils_valid.template operator()<ItemType::node, ItemType::node>(stencil_array, mesh, 2));
            REQUIRE(is_valid.template operator()<ItemType::node, ItemType::node>(connectivity, stencil_array, 2));
          }

          {
            auto stencil_array =
              StencilManager::instance()
                .getNodeToCellStencilArray(connectivity,
                                           StencilDescriptor{2, StencilDescriptor::ConnectionType::by_edges}, list);

            REQUIRE(stencil_array.symmetryBoundaryStencilArrayList().size() == 6);
            REQUIRE(check_ghost_nodes_have_empty_stencils(stencil_array, connectivity));
            REQUIRE(
              are_symmetry_stencils_valid.template operator()<ItemType::node, ItemType::edge>(stencil_array, mesh, 2));
            REQUIRE(is_valid.template operator()<ItemType::node, ItemType::edge>(connectivity, stencil_array, 2));
          }

          {
            auto stencil_array =
              StencilManager::instance()
                .getNodeToCellStencilArray(connectivity,
                                           StencilDescriptor{2, StencilDescriptor::ConnectionType::by_faces}, list);

            REQUIRE(stencil_array.symmetryBoundaryStencilArrayList().size() == 6);
            REQUIRE(check_ghost_nodes_have_empty_stencils(stencil_array, connectivity));
            REQUIRE(
              are_symmetry_stencils_valid.template operator()<ItemType::node, ItemType::face>(stencil_array, mesh, 2));
            REQUIRE(is_valid.template operator()<ItemType::node, ItemType::face>(connectivity, stencil_array, 2));
          }
        }
      }
    }
  }
}
